﻿#include "PianoRollEditTool_Write.h"

#include "Singleton/MidiSequencer.h"
#include "PluginHost/PluginHost.h"
#include "UI/Widget/PianoRoll/PianoRollPanel/SPianoRollPanel.h"

FPianoRollEditTool_Write::FPianoRollEditTool_Write(): FPianoRollEditTool(EPianoRollToolType::Write), Channel(1),
                                                      Velocity(1), NoteLength(FMidiSequencer::Get().GetBeat())
{
	TempSelector.OnMidiMessageSelected.AddRaw(this, &FPianoRollEditTool_Write::OnMidiMessageSelected);
	TempSelector.OnMidiMessageUnSelected.AddRaw(this, &FPianoRollEditTool_Write::OnMidiMessageUnSelected);
}

void FPianoRollEditTool_Write::OnActive(FPianoRollEdit& Edit)
{
	FPianoRollEditTool::OnActive(Edit);
	Edit.Selector.OnMidiMessageSelected.AddRaw(this, &FPianoRollEditTool_Write::OnMidiMessageSelected);
	Edit.Selector.OnMidiMessageUnSelected.AddRaw(this, &FPianoRollEditTool_Write::OnMidiMessageUnSelected);
}

void FPianoRollEditTool_Write::Clear(FPianoRollEdit& Edit)
{
	FPianoRollEditTool::Clear(Edit);
	HasDeleteNote = false;
	TempSelector.UnSelect();
	
	Edit.Selector.OnMidiMessageSelected.RemoveAll(this);
	Edit.Selector.OnMidiMessageUnSelected.RemoveAll(this);
}

void FPianoRollEditTool_Write::OnMouseButtonDown(FPianoRollEdit& Edit, const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	HasDeleteNote = false;
	if (MouseEvent.GetEffectingButton() == EKeys::LeftMouseButton)
	{
		LeftMouseButton(Edit);
	}
	
	if (MouseEvent.GetEffectingButton() == EKeys::RightMouseButton)
	{
		RightMouseButton(Edit);
	}
}

void FPianoRollEditTool_Write::OnMouseButtonUp(FPianoRollEdit& Edit, const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	if (MouseEvent.GetEffectingButton() == EKeys::RightMouseButton && !HasDeleteNote)
		Edit.Selector.UnSelect();
	TempSelector.UnSelect();
}

void FPianoRollEditTool_Write::OnMouseMove(FPianoRollEdit& Edit, const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	const FGeometry& TickSpaceGeometry = MyGeometry;
	const auto& CursorPos = MouseEvent.GetScreenSpacePosition();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(CursorPos);
	const TSharedPtr<SPianoRollPanel> Panel = Edit.PianoRollPanel.Pin();
	
	if (MouseEvent.IsMouseButtonDown(EKeys::LeftMouseButton))
	{
		const int MouseNote = Panel->HeightToNoteNumber(LocalPos.Y);
		const AudioFrame& NewFrame = Panel->PixelToFrame(LocalPos.X);
		UpdateTempSelector(Edit, MouseNote, NewFrame.Ticks());
		UpdateSelector(Edit, MouseNote, NewFrame.Ticks());
	}

	if (MouseEvent.IsMouseButtonDown(EKeys::RightMouseButton))
	{
		RightMouseButton(Edit);
	}
}

void FPianoRollEditTool_Write::OnMidiMessageSelected(FMidiMessageSequence::MidiEventHolder* MidiEventHolder)
{
	PlayNote(MidiEventHolder->message.getNoteNumber());
}

void FPianoRollEditTool_Write::OnMidiMessageUnSelected(FMidiMessageSequence::MidiEventHolder* MidiEventHolder)
{
	StopNote(MidiEventHolder->message.getNoteNumber());
}

void FPianoRollEditTool_Write::PlayNote(int32 Note)
{
	if (GetPluginHost())
		GetPluginHost()->IncomingMidi.addEvent(FMidiMessage::noteOn(Channel, Note, Velocity));
}

void FPianoRollEditTool_Write::StopNote(int32 Note)
{
	if (GetPluginHost())
		GetPluginHost()->IncomingMidi.addEvent(FMidiMessage::noteOff(Channel, Note, Velocity));
}

void FPianoRollEditTool_Write::LeftMouseButton(FPianoRollEdit& Edit)
{
	FPianoRollSelector& Selector = Edit.Selector;
	const MidiTick SnapMidiTimestamp = Edit.GetSnapFrame();
	
	if (const FNoteSelectData& SelectedNote = Edit.GetMidiMessageUnderMouse())
	{
		FMidiMessageSequence::MidiEventHolder* MidiEventHolder = SelectedNote.MidiMessage;
		const int BasePitch = MidiEventHolder->message.getNoteNumber();
		const double MidiTimestampDelta = SelectedNote.MidiTimestampOffset;
		const double BaseMidiTimestamp = MidiEventHolder->message.getTimeStamp();
		Channel = MidiEventHolder->message.getChannel();
		Velocity = MidiEventHolder->message.getFloatVelocity();
		NoteLength = MidiEventHolder->noteOffObject ? MidiEventHolder->noteOffObject->message.getTimeStamp() - BaseMidiTimestamp : 0;
		
		if (Selector.IsSelected(MidiEventHolder))
		{
			Selector.BeginDrag(BasePitch, MidiTimestampDelta, SnapMidiTimestamp, BaseMidiTimestamp);
			Selector.ResetAnchorPoint();
		}
		else
		{
			Selector.UnSelect();
			TempSelector.UnSelect();
			TempSelector.AddMidiMessage(MidiEventHolder);
			TempSelector.BeginDrag(BasePitch, MidiTimestampDelta, SnapMidiTimestamp, BaseMidiTimestamp);
			TempSelector.ResetAnchorPoint();
		}
		return;
	}

	Selector.UnSelect();
	
	const int32 NoteNumberUnderMouse = Edit.GetNoteNumberUnderMouse();
	const MidiTick MidiTickUnderMouse = Edit.GetAudioFrameUnderMouse();
	const MidiTick TimestampDelta = MidiTickUnderMouse % SnapMidiTimestamp;
	const MidiTick SnapMidiTick = MidiTickUnderMouse - TimestampDelta;
	
	const FMidiMessage& NoteOn = FMidiMessage::noteOn(Channel, NoteNumberUnderMouse, Velocity);
	FMidiMessageSequence::MidiEventHolder* MidiEventHolder = Edit.GetMidiMessageSequence()->addEvent(NoteOn, SnapMidiTick);
	if (NoteLength > 0)
		MidiEventHolder->noteOffObject = Edit.GetMidiMessageSequence()->addEvent(FMidiMessage::noteOff(Channel, NoteNumberUnderMouse, Velocity), SnapMidiTick + NoteLength);
}

void FPianoRollEditTool_Write::RightMouseButton(FPianoRollEdit& Edit)
{
	if (const FNoteSelectData& SelectedNote = Edit.GetMidiMessageUnderMouse())
	{
		Edit.Selector.UnSelect(SelectedNote.MidiMessage);
		Edit.SetModified(true);
		FMidiMessageSequence* Sequence = Edit.GetMidiMessageSequence();
		const int NoteIndex = Sequence->getIndexOf(SelectedNote.MidiMessage);
		Sequence->deleteEvent(NoteIndex, true);
		HasDeleteNote = true;
	}
}

void FPianoRollEditTool_Write::UpdateTempSelector(FPianoRollEdit& Edit, int32 NewNote, MidiTick NewTick)
{
	if (TempSelector.Num() == 0)
		return;

	const int32 LastNote = TempSelector.GetSelectedMidiMessages()[0]->message.getNoteNumber();
	bool IsModify = TempSelector.SetPitch(NewNote);
	if (IsModify)
	{
		StopNote(LastNote);
		PlayNote(NewNote);
	}
	
	IsModify |= TempSelector.SetMidiTimestamp(NewTick);
	if (IsModify)
	{
		Edit.SetModified(true);
	}
}

void FPianoRollEditTool_Write::UpdateSelector(FPianoRollEdit& Edit, int32 NewNote, MidiTick NewTick)
{
	if (Edit.Selector.Num() == 0)
		return;

	bool IsModify = Edit.Selector.SetPitch(NewNote);
	IsModify |= Edit.Selector.SetMidiTimestamp(NewTick);
	
	if (IsModify)
	{
		Edit.SetModified(true);
	}
}
