﻿#include "PianoRollEdit.h"

#include "SlateApplication.h"
#include "SPianoRollPanel.h"
#include "Pattern/MidiPattern.h"
#include "Singleton/MidiSequencer.h"
#include "EditTool/PianoRollEditTool.h"
#include "EditTool/PianoRollEditTool_DragView.h"
#include "EditTool/PianoRollEditTool_Select.h"
#include "EditTool/PianoRollEditTool_Write.h"

FPianoRollEdit::FPianoRollEdit(TSharedRef<SPianoRollPanel> InPianoRollPanel)
{
	PianoRollPanel = InPianoRollPanel;
	
	EditToolMap.Add(EPianoRollToolType::Write, new FPianoRollEditTool_Write());
	EditToolMap.Add(EPianoRollToolType::Select, new FPianoRollEditTool_Select());
	EditToolMap.Add(EPianoRollToolType::DragView, new FPianoRollEditTool_DragView());

	
	ToolKeyMap.Add(EPianoRollToolType::Select, {EKeys::LeftControl, EKeys::LeftMouseButton});
	ToolKeyMap.Add(EPianoRollToolType::DragView, {EKeys::LeftShift, EKeys::RightMouseButton});
	

	CurrentEditTool = EditToolMap[EPianoRollToolType::Write];
	CurrentEditTool->OnActive(*this);
	TempTool = nullptr;
}

FPianoRollEdit::~FPianoRollEdit()
{
	for (const auto& Tuple : EditToolMap)
	{
		delete Tuple.Value;
	}
	EditToolMap.Empty();
}

FNoteSelectData FPianoRollEdit::GetMidiMessageByPos(FVector2D Pos) const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	FNoteSelectData OutData;
	
	const int Note = Panel->HeightToNoteNumber(Pos.Y);
	const float& FrameToPixel = Panel->FrameToPixel;
	const TRange<AudioFrame>& FrameRange = Panel->TimeRange;
	FMidiMessageSequence* MidiEventHolders = Panel->MidiMessageSequence;

	// UE_LOG(LogTemp, Log, TEXT("Find Note: %s"), *FMidiMessage::getMidiNoteName(Note, true, true, 3));
	AudioFrame Time = Pos.X / FrameToPixel;
	Time += FrameRange.GetLowerBoundValue();
	const MidiTick Ticks = Time.Ticks();
	for (FMidiMessageSequence::MidiEventHolder* Holder : *MidiEventHolders)
	{
		if (Holder->message.getNoteNumber() != Note)
			continue;
		MidiTick BeginMidiTick = FrameRange.GetLowerBoundValue().Ticks();
		if (Holder->message.getTimeStamp() > BeginMidiTick)
			break;
		if (!Holder->message.isNoteOn())
			continue;
		if (Ticks < Holder->message.getTimeStamp())
			continue;
		
		const FMidiMessageSequence::MidiEventHolder* NoteOffHolder = Holder->noteOffObject;
		if (NoteOffHolder && Ticks > NoteOffHolder->message.getTimeStamp())
			continue;

		OutData.MidiMessage = Holder;
		OutData.MidiTimestampOffset = Ticks - (int64)Holder->message.getTimeStamp();
		return OutData;
	}

	return OutData;
}

FNoteSelectData FPianoRollEdit::GetMidiMessageUnderMouse() const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	const FGeometry& TickSpaceGeometry = Panel->GetTickSpaceGeometry();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(FSlateApplication::Get().GetCursorPos());
	return GetMidiMessageByPos(LocalPos);
}

FReply FPianoRollEdit::OnMouseButtonDown(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	MouseDownPos = MyGeometry.AbsoluteToLocal(MouseEvent.GetScreenSpacePosition());
	TryEnableTempTool();

	GetCurrentTool()->OnMouseButtonDown(*this, MyGeometry, MouseEvent);
	UpdateModified();
	return FReply::Handled();
}

FReply FPianoRollEdit::OnMouseButtonUp(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	const FVector2f MouseUpLocalPos = MyGeometry.AbsoluteToLocal(MouseEvent.GetScreenSpacePosition());
	
	// OnEndDrag需要在OnMouseUp之前调用，因为OnMouseUp可能会清除OnMouseDrag
	if (OnMouseDrag)
	{
		OnMouseDrag = false;
		GetCurrentTool()->OnEndDrag(*this, MouseUpLocalPos);
		UpdateModified();
	}
	
	{
		GetCurrentTool()->OnMouseButtonUp(*this, MyGeometry, MouseEvent);
		UpdateModified();
	}
	
	TryEnableTempTool();

	return FReply::Handled();
}

FReply FPianoRollEdit::OnMouseMove(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	const FGeometry& TickSpaceGeometry = MyGeometry;
	const auto& CursorPos = MouseEvent.GetScreenSpacePosition();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(CursorPos);
	TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	
	if (MouseEvent.IsMouseButtonDown(EKeys::LeftMouseButton) && !OnMouseDrag)
	{
		OnMouseDrag = true;
		GetCurrentTool()->OnBeginDrag(*this, LocalPos);
		UpdateModified();
	}

	GetCurrentTool()->OnMouseMove(*this, MyGeometry, MouseEvent);
	UpdateModified();
	return FReply::Handled();
}

FReply FPianoRollEdit::OnKeyDown(const FGeometry& MyGeometry, const FKeyEvent& KeyEvent)
{
	if (KeyEvent.IsRepeat())
		return FReply::Handled();
	PressedKeys.Add(KeyEvent.GetKey());
	TryEnableTempTool();
	return FReply::Handled();
}

FReply FPianoRollEdit::OnKeyUp(const FGeometry& MyGeometry, const FKeyEvent& KeyEvent)
{
	if (KeyEvent.IsRepeat())
		return FReply::Handled();
	PressedKeys.Remove(KeyEvent.GetKey());
	TryEnableTempTool();
	return FReply::Handled();
}

int32 FPianoRollEdit::OnPaint(const FPaintArgs& Args, const FGeometry& AllottedGeometry,
	const FSlateRect& MyCullingRect, FSlateWindowElementList& OutDrawElements, int32 LayerId,
	const FWidgetStyle& InWidgetStyle, bool bParentEnabled) const
{
	return GetCurrentTool()->OnPaint(*this, Args, AllottedGeometry, MyCullingRect, OutDrawElements, LayerId, InWidgetStyle, bParentEnabled);
}

int32 FPianoRollEdit::GetNoteNumberByPos(float Height) const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	return Panel->HeightToNoteNumber(Height);
}

int32 FPianoRollEdit::GetNoteNumberUnderMouse() const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	const FGeometry& TickSpaceGeometry = Panel->GetTickSpaceGeometry();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(FSlateApplication::Get().GetCursorPos());
	return Panel->HeightToNoteNumber(LocalPos.Y);
}

AudioFrame FPianoRollEdit::GetAudioFrameByPos(float Width) const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	return Panel->PixelToFrame(Width);
}

AudioFrame FPianoRollEdit::GetAudioFrameUnderMouse() const
{
	const TSharedPtr<SPianoRollPanel> Panel = PianoRollPanel.Pin();
	const FGeometry& TickSpaceGeometry = Panel->GetTickSpaceGeometry();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(FSlateApplication::Get().GetCursorPos());
	return Panel->PixelToFrame(LocalPos.X);
}

AudioFrame FPianoRollEdit::GetSnapFrame() const
{
	return FMidiSequencer::Get().GetTicksPerQuarter();
}

FMidiMessageSequence* FPianoRollEdit::GetMidiMessageSequence() const
{
	return PianoRollPanel.Pin()->GetMidiMessageSequence();
}

FPluginHost* FPianoRollEdit::GetPluginHost() const
{
	FMidiMessageSequence* Sequence = GetMidiMessageSequence();
	if (!Sequence)
		return nullptr;
	return Sequence->pattern->GetPluginHost(Sequence);
}

void FPianoRollEdit::SetTempTool(EPianoRollToolType InToolType)
{
	CurrentEditTool->Clear(*this);
	TempTool = EditToolMap[InToolType];
	TempTool->OnActive(*this);
}

void FPianoRollEdit::ClearTempTool()
{
	if (TempTool)
	{
		TempTool->Clear(*this);
		TempTool = nullptr;
		CurrentEditTool->OnActive(*this);
	}
}

void FPianoRollEdit::UpdateModified()
{
	if (GetModified())
	{
		GetMidiMessageSequence()->sortAudioThread();
		SetModified(false);
	}
}

FPianoRollEditTool* FPianoRollEdit::TryEnableTempTool()
{
	const TSet<FKey>& PressedMouseButtons = FSlateApplication::Get().GetPressedMouseButtons();
	for (const auto& Tuple : ToolKeyMap)
	{
		const EPianoRollToolType PianoRollTool = Tuple.Key;
		const TArray<FKey>& Keys = Tuple.Value;
		bool bAllPressed = true;
		for (const FKey& ToolKey : Keys)
		{
			if (ToolKey.IsMouseButton())
				bAllPressed = bAllPressed && PressedMouseButtons.Contains(ToolKey);
			else
				bAllPressed = bAllPressed && PressedKeys.Contains(ToolKey);
			if (!bAllPressed)
				break;
		}
		if (bAllPressed)
		{
			ClearTempTool();
			SetTempTool(PianoRollTool);
			return EditToolMap[PianoRollTool];
		}
	}
	ClearTempTool();
	return nullptr;
}
