﻿// Fill out your copyright notice in the Description page of Project Settings.


#include "SPianoRollPanel.h"

#include "FontMeasure.h"
#include "SImage.h"
#include "SlateApplication.h"
#include "SlateOptMacros.h"
#include "SScaleBox.h"
#include "Singleton/MidiSequencer.h"

BEGIN_SLATE_FUNCTION_BUILD_OPTIMIZATION

void SPianoRollPanel::Construct(const FArguments& InArgs, FMidiMessageSequence* InMidiMessageSequence)
{
	FontInfo = InArgs._FontInfo;
	NoteImage = InArgs._NoteImage;
	MidiMessageSequence = InMidiMessageSequence;
	
	const double MidiDuration = FMath::Min(FMidiSequencer::Get().GetBar() * 2 ,MidiMessageSequence->getEndTime());
	SequenceWidth = MidiDuration * FrameToPixel;
	TimeRange = TRange<AudioFrame>(0, MidiDuration);
	OnTargetRangeChanged = InArgs._OnTargetRangeChanged;
	OnYRequestUpdate = InArgs._OnYRequestUpdate;
	OnRequestHScroll = InArgs._OnRequestHScroll;
	
	SpacingLineColor = InArgs._SpacingLineColor;
	NoteColor = InArgs._NoteColor;
	NoteSelectedColor = InArgs._NoteSelectedColor;
	NoteTextColor = InArgs._NoteTextColor;
	SpacingLineThickness = InArgs._SpacingLineThickness;
	Clipping = EWidgetClipping::ClipToBounds;
	PianoRollEdit = MakeShareable(new FPianoRollEdit(SharedThis(this)));
}

int32 SPianoRollPanel::OnPaint(const FPaintArgs& Args, const FGeometry& AllottedGeometry, const FSlateRect& MyCullingRect,
	FSlateWindowElementList& OutDrawElements, int32 LayerId, const FWidgetStyle& InWidgetStyle,
	bool bParentEnabled) const
{
	const float RangeBegin = TimeRange.GetLowerBoundValue();
	const float RangeEnd = TimeRange.GetUpperBoundValue();
	if (FMath::IsNaN(RangeBegin) || FMath::IsNaN(RangeEnd))
		return LayerId;
	
	const auto& Size = AllottedGeometry.Size;

	const FMidiSequencer& MidiSequencer = FMidiSequencer::Get();
	const double CurrentMidiTick = MidiSequencer.FramePos.load().Ticks();
	const int32 TicksPerQuarter = MidiSequencer.GetTicksPerQuarter();
	int32 TicksPerBeat = MidiSequencer.GetBeat();
	int32 TicksPerBar = MidiSequencer.GetBar();

	++LayerId;
	// 绘制Note间隔线
	for (int32 i = 0; i < FMidiMessage::MaxNoteNumber + 1; ++i)
	{
		TArray<FVector2f> Points;
		Points.Add(FVector2f(0, i * NoteHeight));
		Points.Add(FVector2f(SequenceWidth, i * NoteHeight));
		FSlateDrawElement::MakeLines(
			OutDrawElements,
			LayerId,
			AllottedGeometry.ToPaintGeometry(),
			Points,
			ESlateDrawEffect::None,
			SpacingLineColor,
			true,
			SpacingLineThickness
		);
	}

	++LayerId;
	// 绘制节拍线
	const float Delta = FMath::Fmod(RangeBegin, (float)TicksPerQuarter);
	for (float i = RangeBegin - Delta; i < RangeEnd; i += TicksPerQuarter)
	{
		TArray<FVector2f> Points;
		Points.Add(FVector2f((i - RangeBegin) * FrameToPixel, 0));
		Points.Add(FVector2f((i - RangeBegin) * FrameToPixel, Size.Y));
		FSlateDrawElement::MakeLines(
			OutDrawElements,
			LayerId,
			AllottedGeometry.ToPaintGeometry(),
			Points,
			ESlateDrawEffect::None,
			SpacingLineColor,
			true,
			SpacingLineThickness
		);
	}

	++LayerId;
	// 绘制Note
	{
		for (const auto& MidiMessage : *MidiMessageSequence)
		{
			if (!MidiMessage->message.isNoteOn())
				continue;
			double NoteOnTime = MidiMessage->message.getTimeStamp();
			if (RangeEnd < NoteOnTime)
				break;

			float NoteOffRealTime;
			if (MidiMessage->noteOffObject)
			{
				NoteOffRealTime = MidiMessage->noteOffObject->message.getTimeStamp();
				if (RangeBegin > NoteOffRealTime)
					continue;
				NoteOffRealTime -= RangeBegin;
			}
			else
			{
				NoteOffRealTime = 0;
				if (RangeBegin > NoteOnTime)
					continue;
			}

			NoteOnTime -= RangeBegin;
		
			const int NoteNumber = MidiMessage->message.getNoteNumber();
			const float NoteOnX = NoteOnTime * FrameToPixel;
			const float NoteOffX = NoteOffRealTime * FrameToPixel;
			const float NoteY = (FMidiMessage::MaxNoteNumber - NoteNumber) * NoteHeight;
			const float NoteWidth = NoteOffX - NoteOnX;
			const FSlateLayoutTransform& InLayoutTransform =  FSlateLayoutTransform(FVector2D(NoteOnX, NoteY));
			// 根据NoteHeight绘制矩形
			FSlateDrawElement::MakeBox(
				OutDrawElements,
				LayerId,
				AllottedGeometry.ToPaintGeometry(FVector2D(NoteWidth, NoteHeight), InLayoutTransform),
				NoteImage,
				ESlateDrawEffect::None,
				PianoRollEdit->IsNoteSelected(MidiMessage) ? NoteSelectedColor : NoteColor
			);
		
			const FString& NoteName = FMidiMessage::getMidiNoteName(NoteNumber, true, true);
		
			// 绘制Note名称
			FSlateDrawElement::MakeText(
				OutDrawElements,
				LayerId,
				AllottedGeometry.ToPaintGeometry(FVector2D(NoteWidth, NoteHeight), InLayoutTransform),
				NoteName,
				FontInfo,
				ESlateDrawEffect::None,
				NoteTextColor
			);
		}
	}

	++LayerId;
	// 绘制当前MidiTick
	{
		const float CurrentMidiTickX = (CurrentMidiTick - RangeBegin) * FrameToPixel;
		TArray<FVector2f> Points;
		Points.Add(FVector2f(CurrentMidiTickX, 0));
		Points.Add(FVector2f(CurrentMidiTickX, Size.Y));
		FSlateDrawElement::MakeLines(
			OutDrawElements,
			LayerId,
			AllottedGeometry.ToPaintGeometry(),
			Points,
			ESlateDrawEffect::None,
			FLinearColor::Red,
			true,
			1
		);
	}

	LayerId = PianoRollEdit->OnPaint(Args, AllottedGeometry, MyCullingRect, OutDrawElements, LayerId, InWidgetStyle, bParentEnabled);
	
	return LayerId;
}

FReply SPianoRollPanel::OnMouseButtonDown(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	return PianoRollEdit->OnMouseButtonDown(MyGeometry, MouseEvent);
}

FReply SPianoRollPanel::OnMouseButtonUp(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	return PianoRollEdit->OnMouseButtonUp(MyGeometry, MouseEvent);
}

FReply SPianoRollPanel::OnMouseMove(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	return PianoRollEdit->OnMouseMove(MyGeometry, MouseEvent);
}

FReply SPianoRollPanel::OnKeyDown(const FGeometry& MyGeometry, const FKeyEvent& InKeyEvent)
{
	if (!InKeyEvent.IsRepeat() && InKeyEvent.GetKey() == EKeys::LeftShift)
		OnRequestHScroll.ExecuteIfBound(true);
	return PianoRollEdit->OnKeyDown(MyGeometry, InKeyEvent);
}

FReply SPianoRollPanel::OnKeyUp(const FGeometry& MyGeometry, const FKeyEvent& InKeyEvent)
{
	if (!InKeyEvent.IsRepeat() && InKeyEvent.GetKey() == EKeys::LeftShift)
		OnRequestHScroll.ExecuteIfBound(false);
	
	return PianoRollEdit->OnKeyUp(MyGeometry, InKeyEvent);
}

END_SLATE_FUNCTION_BUILD_OPTIMIZATION
