﻿#include "PlayListEdit.h"

#include "SlateApplication.h"
#include "Pattern/PatternInstance.h"
#include "EditTool/PlayListEditTool_DragView.h"
#include "EditTool/PlayListEditTool_Select.h"
#include "EditTool/PlayListEditTool_Write.h"
#include "UI/Widget/Pattern/SPatternInstance.h"
#include "UI/Widget/Pattern/SSamplePatternInstance.h"
#include "UI/Widget/PlayList/PlayListPanel/SPlayListPanel.h"

FPlayListEdit::FPlayListEdit(TSharedRef<SPlayListPanel> InPlayListPanel)
{
	PlayListPanel = InPlayListPanel;

	EditToolMap.Add(EPlayListToolType::Select, new FPlayListEditTool_Select());
	EditToolMap.Add(EPlayListToolType::Write, new FPlayListEditTool_Write());
	EditToolMap.Add(EPlayListToolType::DragView, new FPlayListEditTool_DragView());

	ToolKeyMap.Add(EPlayListToolType::Select, { EKeys::LeftControl, EKeys::LeftMouseButton });
	ToolKeyMap.Add(EPlayListToolType::DragView, { EKeys::LeftShift, EKeys::RightMouseButton });
	
	CurrentEditTool = EditToolMap[EPlayListToolType::Write];
	CurrentEditTool->OnActive(*this);
	TempTool = nullptr;
}

FReply FPlayListEdit::OnMouseButtonDown(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	MouseDownPos = MyGeometry.AbsoluteToLocal(MouseEvent.GetScreenSpacePosition());
	TryEnableTempTool();

	GetCurrentTool()->OnMouseButtonDown(*this, MyGeometry, MouseEvent);
	return FReply::Handled();
}

FReply FPlayListEdit::OnMouseButtonUp(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	const FVector2f MouseUpLocalPos = MyGeometry.AbsoluteToLocal(MouseEvent.GetScreenSpacePosition());

	// OnEndDrag需要在OnMouseUp之前调用，因为OnMouseUp可能会清除OnMouseDrag
	if (OnMouseDrag)
	{
		OnMouseDrag = false;
		GetCurrentTool()->OnEndDrag(*this, MouseUpLocalPos);
	}
	
	{
		GetCurrentTool()->OnMouseButtonUp(*this, MyGeometry, MouseEvent);
	}
	
	TryEnableTempTool();
	return FReply::Handled();
}

FReply FPlayListEdit::OnMouseMove(const FGeometry& MyGeometry, const FPointerEvent& MouseEvent)
{
	const FGeometry& TickSpaceGeometry = MyGeometry;
	const auto& CursorPos = MouseEvent.GetScreenSpacePosition();
	const auto& LocalPos = TickSpaceGeometry.AbsoluteToLocal(CursorPos);
	
	if (MouseEvent.IsMouseButtonDown(EKeys::LeftMouseButton) && !OnMouseDrag)
	{
		OnMouseDrag = true;
		GetCurrentTool()->OnBeginDrag(*this, LocalPos);
	}

	GetCurrentTool()->OnMouseMove(*this, MyGeometry, MouseEvent);
	return FReply::Handled();
}

FReply FPlayListEdit::OnKeyDown(const FGeometry& MyGeometry, const FKeyEvent& InKeyEvent)
{
	if (InKeyEvent.IsRepeat())
		return FReply::Handled();
	PressedKeys.Add(InKeyEvent.GetKey());
	TryEnableTempTool();
	return FReply::Handled();
}

FReply FPlayListEdit::OnKeyUp(const FGeometry& MyGeometry, const FKeyEvent& InKeyEvent)
{
	if (InKeyEvent.IsRepeat())
		return FReply::Handled();
	PressedKeys.Remove(InKeyEvent.GetKey());
	TryEnableTempTool();
	return FReply::Handled();
}

int32 FPlayListEdit::OnPaint(const FPaintArgs& Args, const FGeometry& AllottedGeometry, const FSlateRect& MyCullingRect,
	FSlateWindowElementList& OutDrawElements, int32 LayerId, const FWidgetStyle& InWidgetStyle,
	bool bParentEnabled) const
{
	return GetCurrentTool()->OnPaint(*this, Args, AllottedGeometry, MyCullingRect, OutDrawElements, LayerId, InWidgetStyle, bParentEnabled);
}

FPatternInstanceSelectData FPlayListEdit::GetPatternInstanceAt(const FVector2f& LocalPosition) const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();

	FPatternInstanceSelectData Result;
	
	const int32 TrackIndex = Panel->TrackHeightManager.GetTrackIndexAt(LocalPosition.Y);
	const AudioFrame FrameTimestamp = Panel->PixelToFrame(LocalPosition.X);
	
	FMidiSequencer::Get().ForeachPatternInstance([&](FPatternInstance* Instance) ->bool
	{
		if (Instance->TrackIndex != TrackIndex)
			return true;
		const AudioFrame InstancePos = Instance->Pos;
		const AudioFrame InstanceLength = Instance->GetLength();
		
		if (FrameTimestamp >= InstancePos && FrameTimestamp <= InstancePos + InstanceLength)
		{
			Result.PatternInstance = Instance;
			Result.MidiTimestampOffset = FrameTimestamp - InstancePos;
			return false;
		}
		return true;
	});
	return Result;
}

FPatternInstanceSelectData FPlayListEdit::GetPatternUnderCursor() const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();
	const FVector2f& CursorPos = FSlateApplication::Get().GetCursorPos();
	const FVector2f& LocalPosition = Panel->GetTickSpaceGeometry().AbsoluteToLocal(CursorPos);
	return GetPatternInstanceAt(LocalPosition);
}

FReply FPlayListEdit::OnPatternInstancePressed(FPatternInstanceClickData D)
{
	if (!IsMultiSelected())
		Selector.UnSelect();
	else
		return FReply::Unhandled();
	
	const int32 BaseTrackIndex = D.PatternInstance->TrackIndex;
	const double MidiTimestampDelta = D.MidiTimestampOffset;
	const double BaseMidiTimestamp = D.PatternInstance->Pos;
	
	Selector.BeginDrag(BaseTrackIndex, GetSnapFrame(), MidiTimestampDelta, BaseMidiTimestamp);
	Selector.AddPatternInstance(D.PatternInstance);
	Selector.ResetAnchorPoint();
	IsPatternInstancePressed = true;
	return FReply::Handled().CaptureMouse(PlayListPanel.Pin().ToSharedRef());
}

FReply FPlayListEdit::OnPatternInstanceReleased(FPatternInstanceClickData D)
{
	IsPatternInstancePressed = false;
	Selector.EndSetting();
	return FReply::Handled().ReleaseMouseCapture();
}

TSharedRef<SPatternInstance> FPlayListEdit::CreatePatternInstanceWidget(FPatternInstance* PatternInstance)
{
	TSharedPtr<SPatternInstance> PatternInstanceWidget;
	
	switch (PatternInstance->GetOwner()->Type)
	{
	case EPatternType::Midi:
		break;
	case EPatternType::Automation:
		break;
	case EPatternType::Sample:
		PatternInstanceWidget = SNew(SSamplePatternInstance)
			.PatternInstance(PatternInstance)
			.FrameViewRange(PlayListPanel.Pin()->GetViewRange())
			.FrameToPixel_Raw(this, &FPlayListEdit::GetSampleToPixelScaler)
			.SnapFrame_Raw(this, &FPlayListEdit::GetSnapFrame)
			.Clipping(EWidgetClipping::ClipToBoundsAlways);
		break;
	}

	return PatternInstanceWidget.ToSharedRef();
}

AudioFrame FPlayListEdit::GetSnapFrame() const
{
	return FMidiSequencer::Get().GetBeat().Frames();
}

float FPlayListEdit::GetMidiTickToPixelScaler() const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();
	return Panel->GetMidiTickToPixelScaler();
}

float FPlayListEdit::GetSampleToPixelScaler() const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();
	return Panel->GetSampleToPixelScaler();
}

AudioFrame FPlayListEdit::GetFrameUnderCursor() const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();
	const FVector2f& CursorPos = FSlateApplication::Get().GetCursorPos();
	const FVector2f& LocalPosition = Panel->GetTickSpaceGeometry().AbsoluteToLocal(CursorPos);
	return Panel->PixelToFrame(LocalPosition.X);
}

AudioFrame FPlayListEdit::PixelToFrame(float Pixel) const
{
	return PlayListPanel.Pin()->PixelToFrame(Pixel);
}

int32 FPlayListEdit::GetTrackIndexUnderCursor() const
{
	const TSharedPtr<SPlayListPanel> Panel = PlayListPanel.Pin();
	const FVector2f& CursorPos = FSlateApplication::Get().GetCursorPos();
	const FVector2f& LocalPosition = Panel->GetTickSpaceGeometry().AbsoluteToLocal(CursorPos);
	return Panel->TrackHeightManager.GetTrackIndexAt(LocalPosition.Y);
}

int32 FPlayListEdit::GetTrackIndexByPos(float Height) const
{
	return PlayListPanel.Pin()->TrackHeightManager.GetTrackIndexAt(Height);
}

void FPlayListEdit::SetTempTool(EPlayListToolType InToolType)
{
	CurrentEditTool->Clear(*this);
	TempTool = EditToolMap[InToolType];
	TempTool->OnActive(*this);
}

void FPlayListEdit::ClearTempTool()
{
	if (TempTool)
	{
		TempTool->Clear(*this);
		TempTool = nullptr;
		CurrentEditTool->OnActive(*this);
	}
}

FPlayListEditTool* FPlayListEdit::TryEnableTempTool()
{
	const TSet<FKey>& PressedMouseButtons = FSlateApplication::Get().GetPressedMouseButtons();
	for (const auto& Tuple : ToolKeyMap)
	{
		const EPlayListToolType PianoRollTool = Tuple.Key;
		const TArray<FKey>& Keys = Tuple.Value;
		bool bAllPressed = true;
		for (const FKey& ToolKey : Keys)
		{
			if (ToolKey.IsMouseButton())
				bAllPressed = bAllPressed && PressedMouseButtons.Contains(ToolKey);
			else
				bAllPressed = bAllPressed && PressedKeys.Contains(ToolKey);
			if (!bAllPressed)
				break;
		}
		if (bAllPressed)
		{
			ClearTempTool();
			SetTempTool(PianoRollTool);
			return EditToolMap[PianoRollTool];
		}
	}
	ClearTempTool();
	return nullptr;
}
